/* Copyright (C) 2006 - Kimball Group - All Rights Reserved  */
/* Version 3.81 - August 16, 2006                            */
/* This script performs the initial population of the        */
/* DW_Metadata schema based on the use of extended           */
/* properties and naming conventions described in the        */
/* Microsoft Data Warehouse Toolkit book and implemented     */
/* in the associated dimensional data modeling spread-       */
/* sheet (which you can obtain from the book's website at    */
/* http://www.msftdwtoolkit.com).                            */
/* Usage:
   First, create the DW_Metadata database and associated  
   tables using the "Metadata database and table create script.sql" file. 
   Then run this script from the metadata catalog database. 
  
   After the first run, you will see an entry for each 
   database on the server, most of which do not have metadata
   extended properties. If you would like to remove any,
   of these databases, edit the DW_Databases table and set  
   the Gen_Metadata_Ind field to 'N' for each database you 
   would like to remove.
 
   Then, run this script again and it will clear out the
   first run and repopulate the metadata tables based on 
   the extended properties as described above.

   You can browse the metadata tables using the Reporting
   Package from the Chapter 13 Tools and Utilities on the 
   book's website.
*/

use DW_Metadata
Declare @Debug bit
  SET @Debug = 'False'

IF @Debug = 'False' SET NOCOUNT ON

Create Table #DBDescr (
  DatabaseID int,
  Class Int,
  Class_Desc varchar(256),
  Major_ID int,
  Minor_ID int,
  ExtPropName nvarchar(1000),
  ExtPropValue nvarchar(1000)
)

DECLARE
	@DBID int,
	@DBIDString sysname,
	@DBName sysname,
	@DatabaseKey int,
	@ParmDefinition nvarchar(500)

SET @ParmDefinition = N'@DatabaseName sysname, @DatabaseID int'

DECLARE DBCursor CURSOR FOR 
SELECT database_id, name from sys.databases
where owner_sid <> 1

OPEN DBCursor

FETCH NEXT FROM DBCursor 
INTO   @DBID, @DBName

WHILE @@FETCH_STATUS = 0
BEGIN
     EXEC sp_executesql N'EXECUTE (N''INSERT INTO #DBDescr SELECT '''''' + @DatabaseID + '''''' DBID, class, class_desc, major_id, 
     minor_id, CAST(name AS nvarchar(1000)), CAST(value AS nvarchar(1000)) from ['' + @DatabaseName + ''].sys.extended_properties
	 WHERE Class=0 AND name like ''''%Description%'''' '')', @ParmDefinition, @DatabaseName = @DBName, @DatabaseID = @DBID
	-- Could put the database-specific queries in here...

	FETCH NEXT FROM DBCursor 
	INTO   @DBID, @DBName
END  
CLOSE DBCursor
DEALLOCATE DBCursor

-- Clear out the metadata already in the metadata tables
DELETE FROM DW_Metadata.dbo.DW_Subject_Area_Contents
DELETE FROM DW_Metadata.dbo.DW_Object_Contents
DELETE FROM DW_Metadata.dbo.DW_Database_Contents
DELETE FROM DW_Metadata.dbo.DW_Subject_Areas
DELETE FROM DW_Metadata.dbo.DW_Objects
TRUNCATE TABLE DW_Metadata.dbo.DW_Subject_Area_Contents
DELETE FROM  DW_Metadata.dbo.DW_ATTRIBUTEs


-- Insert any new databases into the Database table
-- delete from DW_Metadata.dbo.DW_Databases  -- if you really want to start over, uncomment this line
INSERT INTO DW_Databases
(Database_ID, Database_Name, Database_Short_Descr, Database_Long_Descr, 
 Gen_Metadata_Ind, Database_Type, Database_Status, Instance_Name, Server_Name)
SELECT 
database_id, a.name, 
ExtPropValue ShortDescr,
ExtPropValue LongDescr,
'Y' Gen_Metadata_Ind,  -- default is to generate metadata. Change this to N for databases that don't have metadata structures.
'User' Database_Type, -- We are only loading User databases
'Active' Database_Status,
CAST(SERVERPROPERTY('InstanceName') AS Varchar) Instance_Name,
CAST(SERVERPROPERTY('ServerName') AS Varchar) Server_Name
FROM sys.databases a
LEFT OUTER JOIN #DBDescr b
ON a.database_id = b.DatabaseID
WHERE owner_sid <> 1 -- non-system databases
AND a.database_id NOT IN (select Database_ID from DW_Databases) -- Only new databases


/* Create Subject Areas for a given database.                                  */
/* We used an Identity column for the Subject_Area_ID to accomodate a range of */
/* subject areas, but that means we have to populate the Subject_Area table first  */
/* in order to have Subject_Area_IDs to use in the Subject_Area_Contents table.  */
/* We'll have to keep mapping back to the Subject_Area table to get the    */
/* current Subject Area ID */
/* We identify subject areas based on naming conventions for the central   */
/* Fact tables.  If there is an extended property called Table Type with   */
/* the value of Fact, we use that as the core table in a subject area.   */

DECLARE
	@LoopCounter int,
	@MaxDBs int,
	@SQLString varchar(4000)

SET @LoopCounter = (SELECT MIN(Database_Key) from DW_Metadata.dbo.DW_Databases)
SET @MaxDBs = (SELECT MAX(Database_Key) from DW_Metadata.dbo.DW_Databases WHERE Gen_Metadata_Ind='Y')

WHILE @LoopCounter <= @MaxDBs
BEGIN
-- Find the next database marked for metadata generation
SET @LoopCounter = (SELECT MIN(Database_Key) FROM DW_Metadata.dbo.DW_Databases 
WHERE Database_Key >= @LoopCounter AND Gen_Metadata_Ind='Y')

-- Load the database variables
SELECT @DatabaseKey = Database_Key, @DBID = Database_ID, @DBName = Database_Name 
FROM DW_Databases 
WHERE Database_Key = @LoopCounter

SET @DBIDString = CAST(@DBID AS sysname)

/* Fill in the subject areas by finding the fact tables.  Every fact table */
/* gets its own subject area as a starting point.  */

SET @SQLString = 'INSERT INTO DW_Metadata.dbo.DW_Subject_Areas 
(Database_ID, Primary_Fact_Table_ID, Subject_Area_Name, Subject_Area_Short_Descr, 
Subject_Area_Long_Descr, Business_Process_Area, Subject_Area_Status, Subject_Area_Image)
SELECT DISTINCT ' + @DBIDString + ', a.object_id Primary_Fact_Table_ID, 
CASE WHEN UPPER(LEFT(a.name, 4)) = ''FACT'' THEN RIGHT(a.name, LEN(a.name)-4) ELSE a.name END SubjectAreaName, 
''The '' + CASE WHEN UPPER(LEFT(a.name, 4)) = ''FACT'' THEN RIGHT(a.name, LEN(a.name)-4) ELSE a.name END + '' Subject Area is about '' + CASE WHEN UPPER(LEFT(a.name, 4)) = ''FACT'' THEN RIGHT(a.name, LEN(a.name)-4) ELSE a.name END Short_Descr, 
''The '' + CASE WHEN UPPER(LEFT(a.name, 4)) = ''FACT'' THEN RIGHT(a.name, LEN(a.name)-4) ELSE a.name END + '' Subject Area is totally about '' + CASE WHEN UPPER(LEFT(a.name, 4)) = ''FACT'' THEN RIGHT(a.name, LEN(a.name)-4) ELSE a.name END Long_Descr, 
''Sales'' Business_Process_Area, ''Active'' SubjectAreaStatus, Null SubjectAreaImage
from ' 
+ @DBName + '.sys.objects a
join ' + @DBName + '.sys.extended_properties b
on a.object_id = b.major_id
where b.name = ''Table Type'' and b.value IN (''Fact'')'
EXECUTE (@SQLString)

IF @Debug = 'TRUE' SELECT @SQLString

/* Do the following Subject Areas separately depending on which database  */
/* contains the associated tables.  For example, the utility tables could be in an ETL   */
/* database on a different server.   */

/* Create a Utility subject area if there are any tables with a Utility table type */
SET @SQLString = 'INSERT INTO DW_Metadata.dbo.DW_Subject_Areas 
(Database_ID, Primary_Fact_Table_ID, Subject_Area_Name, Subject_Area_Short_Descr, 
Subject_Area_Long_Descr, Business_Process_Area, Subject_Area_Status, Subject_Area_Image)
SELECT ' + CAST(@DBID AS Varchar(50)) + ', NULL, ''Utility'' SubjectAreaName, 
''Utility tables are used to build the final Fact and Dimension tables.'' Short_Descr, 
''The Utility tables are typically lookup tables or mapping tables that the ETL process uses to create the user queryable star schema. '' + 
''They are not usually made available for direct user query. '' Long_Descr, 
''Data Warehouse'' Bus_Process_Area, ''Active'' SubjectAreaStatus, NULL SubjectAreaImage
WHERE EXISTS (SELECT Name FROM ' + @DBName + '.sys.extended_properties 
WHERE Name = ''Table Type'' and Value = ''Utility'')'
--SELECT @SQLString
EXECUTE (@SQLString)

/* Create an Audit subject area if there are any tables with an Audit table type */
SET @SQLString = 'INSERT INTO DW_Metadata.dbo.DW_Subject_Areas 
(Database_ID, Primary_Fact_Table_ID, Subject_Area_Name, Subject_Area_Short_Descr, 
Subject_Area_Long_Descr, Business_Process_Area, Subject_Area_Status, Subject_Area_Image)
SELECT ' + CAST(@DBID AS Varchar(50)) + ', NULL, ''Audit'' SubjectAreaName, 
''Audit tables are used to track the ETL process and to capture specific data quality measures.'' Short_Descr, 
''The ETL process writes out entries to the Audit tables which can then be used to identify the load process that was responsible for a given row in the star schema.'' + 
''The Audit tables also capture process metadata such as start times and stop times for the various steps in the ETL job stream.'' Long_Descr, 
''Data Warehouse'' Bus_Process_Area, ''Active'' SubjectAreaStatus, NULL SubjectAreaImage
WHERE EXISTS (SELECT Name FROM ' + @DBName + '.sys.extended_properties 
WHERE Name = ''Table Type'' and Value = ''Audit'')'

EXECUTE (@SQLString)

IF @Debug = 'TRUE' PRINT 'At end of Subject_Area section for ' + @DBName

/* Map the Databases to the Subject Areas in the DW_Database_Contents table */
Insert into DW_Metadata.dbo.DW_Database_Contents
Select
Database_Key, Subject_Area_Key 
FROM
DW_Metadata.dbo.DW_Databases D,
DW_Metadata.dbo.DW_Subject_Areas SA
WHERE D.Database_Key = @DatabaseKey 
AND D.Database_ID = SA.Database_ID

IF @Debug = 'TRUE' PRINT 'At end of Database_Contents section for ' + @DBName

/* Insert all the Objects in to the DW_Objects table.  Even   */
/* though some of them are not contained in the DW_Object_Content  */
/* table.  This will allow you to add the un-mapped objects to  */
/* the appropriate subject areas later.                        
   Note that you need to add role playing tables once for each role 
   they play.  The Subject_Area_Contents table will need to map 
   the different roles to the subject area. 
*/

SET @SQLString = 'INSERT INTO DW_Metadata.dbo.DW_Objects
SELECT
a.object_id Object_ID, ' + @DBIDString + ',
a.name Object_Display_Name, -- initial population only from RDBMS - do not overwrite on update
a.name Object_Physical_Name,
''Table'' Object_Type,
CONVERT(varchar(100),MAX(CASE WHEN b.name=''Table Type'' THEN b.value ELSE '''' END)) Object_Purpose,
CONVERT(varchar(255),MAX(CASE WHEN b.name=''Description'' THEN b.value ELSE '''' END)) ShortDescription,
CONVERT(varchar(2000),MAX(CASE WHEN b.name=''Description'' THEN b.value ELSE '''' END)) LongDescription, -- initial population only from RDBMS - do not overwrite on update
''To be filled in...'' Object_Usage_Tips, -- initial load only - do not overwrite
''To be filled in...'' Object_Steward_Contact, -- Initial load only - do not overwrite
getdate() Last_refresh_date,
''Daily'' Update_Frequency,
''Active'' Object_Status
FROM ' + @DBName + '.sys.objects a
LEFT OUTER JOIN ' + @DBName + '.sys.extended_properties b  -- to find the Tables in each Subject Area
ON b.Major_ID = a.object_id
where b.minor_id = 0
--where A.type = ''U''
GROUP BY a.object_id, a.name'
EXECUTE (@SQLString)
IF @Debug = 'TRUE' PRINT 'At end of Objects section for ' + @DBName

/* Load the Subject Area Contents table based on 'FK To' extended property  
   from the original spreadsheet model.  This section starts with fact tables 
   as the root because it is based on the existence of the Primary_Fact_Table_ID 
   in the DW_Subject_Area table.  */

-- Start with the fact tables
SET @SQLString = 'USE ' + @DBName + ' INSERT INTO DW_Metadata.dbo.DW_Subject_Area_Contents
select Subject_Area_Key, Object_Key, O.Object_Physical_Name 
from DW_Metadata.dbo.DW_Subject_Areas SA JOIN 
DW_Metadata.dbo.DW_Objects O on
SA.Database_ID = O.Database_ID AND
SA.Primary_Fact_Table_ID = O.Object_ID
WHERE SA.Database_ID = ' + CAST(@DBID AS Varchar(50))
IF @Debug = 'TRUE' select @SQLString
EXECUTE(@SQLString)

-- Now get the associated dimension tables
SET @SQLString = 'USE ' + @DBName + ' INSERT INTO DW_Metadata.dbo.DW_Subject_Area_Contents
SELECT d.Subject_Area_Key, c.Object_Key, cols.name
FROM sys.objects a  /* subject area */
JOIN sys.extended_properties b  /* to find the FK columns */
on a.object_id = b.major_id
JOIN DW_Metadata.dbo.DW_Objects c  /* to look up the FK objects in a subject area */
ON (LEFT(CAST(b.value AS VARCHAR(2000)), ABS(CHARINDEX(''.'',CAST(b.value AS VARCHAR(2000)))-1))) = c.Object_Physical_Name  
  /* have to join by name because that is all we have in the extended properties */
JOIN DW_Metadata.dbo.DW_Subject_Areas d
ON a.object_ID = d.Primary_Fact_Table_ID
JOIN sys.columns cols on a.object_ID = cols.object_ID
   AND cols.column_ID = b.minor_id
where b.name = ''FK To''
AND c.Database_ID = d.Database_ID
AND d.Database_ID = ' + CAST(@DBID As varchar(250)) + '
order by 1'

IF @Debug = 'TRUE' select @SQLString
EXECUTE(@SQLString)

/* Insert the Audit and Utility tables into the DW_Subject_Area_Content.    
   NOTE: This code will miss any Audit or Utility tables that are marked   
   as Dimensions in the extended properties.  If they are not really   
   dimensions, in other words, if they are not foreign keys to a Fact table,
   you should set their Table Type to Utility if you want them to show up
   in the Metadata database.
*/

SET @SQLString = 'USE ' + @DBName + ' INSERT INTO 
 DW_Metadata.dbo.DW_Subject_Area_Contents
SELECT DISTINCT d.Subject_Area_Key, c.Object_Key, c.Object_Physical_Name
FROM sys.extended_properties b  -- to find the Tables in each Subject Area
JOIN DW_Metadata.dbo.DW_Subject_Areas d
ON b.value = d.Subject_Area_Name
JOIN DW_Metadata.dbo.DW_Objects c
ON c.Object_ID = b.Major_ID
   AND c.Database_ID = d.Database_ID
where 
d.Database_ID = ' + CAST(@DBID As varchar(250)) + '
AND b.name = ''Table Type''
AND b.value IN (''Audit'', ''Utility'', ''Maintenance'', ''Bridge'') -- add any other table types as needed
order by 1'

EXECUTE(@SQLString)

IF @Debug = 'TRUE' PRINT 'At end of Subject_Area_Contents for ' + @DBName

/* NOTE: This code does not include Bridge tables and their associated  
 dimensions in the appropriate subject areas.  You will need to manually 
 remap them in the Metadata Database.
*/

/* Now set the Join_Column to a zero length string if the relationship is not a role-
   playing relationship.  You can then concatenate the Object Name with its Join_Column
   in the reporting process, although you might as well put a display name column in 
   every table that would pre-join these columns.  
*/

Update DW_Subject_Area_Contents
  Set Join_Column = '' 
  From (Select Subject_Area_Key, Object_Key, count(*) As Row_Count 
        from DW_Subject_Area_Contents Group BY Subject_Area_Key, Object_Key) Non_Role_Tables
  JOIN DW_Subject_Area_Contents SAC
    ON SAC.Subject_Area_Key = Non_Role_Tables.Subject_Area_Key
    AND SAC.Object_Key = Non_Role_Tables.Object_Key
  AND Row_Count=1

IF @Debug = 'TRUE' select * from DW_Metadata.dbo.DW_Subject_Area_Contents

/* Populate the Attributes table so the Object Contents table  */
/* won't fail RI checks when we insert  */

/* Add in all the Attributes for the objects in the current database */
SET @SQLString = 'USE ' + @DBName + ' INSERT INTO 
 DW_Metadata.dbo.DW_Attributes
SELECT
(''' + CAST(@DBID AS varchar(50)) + ''' + ''-'' + CAST(c.object_id AS varchar(30)) + ''-'' + CAST(c.column_id AS varchar(30))) Attribute_ID, 
' + CAST(@DBID AS varchar(50)) + ' Database_ID, a.object_id Object_ID, c.column_id ColID, 
CONVERT(varchar(2000),MAX(CASE WHEN b.name=''Display Name'' THEN b.value ELSE '''' END)) AS [Attribute Display Name],
c.name Attribute_Physical_Name,
''Default'' Attribute_Group_Name,
d.Data_Type,
ISNULL(Character_Maximum_Length,0) + ISNULL(Numeric_Precision,0) + ISNULL(DATETIME_PRECISION,0) Length, 
CONVERT(varchar(255),MAX(CASE WHEN b.name=''Description'' THEN b.value ELSE '''' END)) ShortDescription,
CONVERT(varchar(2000),MAX(CASE WHEN b.name=''Description'' THEN b.value ELSE '''' END)) LongDescription, -- initial population only from RDBMS - do not overwrite on update
CONVERT(varchar(512),MAX(CASE WHEN b.name=''Example Values'' THEN b.value ELSE '''' END)) AS [Example Values],
CONVERT(varchar(2000),MAX(CASE WHEN b.name=''Comment'' THEN b.value ELSE '''' END)) AS [Comment],
d.COLUMN_DEFAULT DefaultValue,
CONVERT(varchar(2000),MAX(CASE WHEN b.name=''ETL Rules'' THEN b.value ELSE '''' END)) AS [ETL Rules],
CONVERT(varchar(255),MAX(CASE WHEN b.name=''Source System'' THEN b.value ELSE '''' END)) AS [Source System],
''To be filled in...'' Attribute_Usage_Tips, -- initial load only - do not overwrite
''To be filled in...'' Attribute_Steward_Contact, -- Initial load only - do not overwrite
''Active'' Attribute_Status
from sys.tables a
JOIN sys.columns c
on a.object_id = c.object_id 
JOIN sys.schemas e
ON a.schema_id = e.schema_id
JOIN INFORMATION_SCHEMA.COLUMNS d
ON d.COLUMN_NAME = c.name
AND d.TABLE_NAME = a.name
AND d.TABLE_SCHEMA = e.name
LEFT OUTER JOIN sys.extended_properties b
on a.object_id = b.major_id
and b.minor_id = c.column_id
group by a.object_id, a.name, c.object_id, c.column_id, c.name, 
d.DATA_TYPE, ISNULL(Character_Maximum_Length,0) + ISNULL(Numeric_Precision,0) + ISNULL(DATETIME_PRECISION,0),
d.COLUMN_DEFAULT
order by a.name, c.column_id'
IF @Debug = 'TRUE' SELECT @SQLString
EXECUTE(@SQLString)

SET @LoopCounter = @LoopCounter + 1
END -- Database Loop

/* Set all the Attribute Display Names to be the same as the physical names if the display name is blank */

Update DW_Metadata.dbo.DW_Attributes set Attribute_Display_Name = Attribute_Physical_Name
       WHERE LEN(Attribute_Display_Name) = 0

/* 
Finally, populate the Object_Content info to map Attributes to Objects.  
The schema allows for multiple mappings.  This code loads all 
attributes for each object into the Default mapping.  Note that 
there are probably some objects that are not in the             
Subject_Area_Content table.  This allows you to use the mapping  
tool you will create to add new Subject Areas and map the objects 
to them.  Please keep us posted of your progress in creating this tool. 
*/

INSERT INTO DW_Metadata.dbo.DW_Object_Contents
SELECT
a.Object_Key, c.Attribute_Key, 
'Default' Content_Map_Name
from DW_Metadata.dbo.DW_Objects a
JOIN DW_Metadata.dbo.DW_Attributes c
on a.object_id = c.object_id 
   AND a.Database_ID = c.Database_ID


-- Clean up
DROP TABLE #DBDescr

IF @Debug = 'TRUE' 
BEGIN
  Select * from DW_Databases
  Select * from DW_Subject_Areas
  Select * from DW_Database_Contents
  Select * from DW_Objects
  Select Subject_Area_Key, Object_Key, count(*) from DW_Subject_Area_Contents group by Subject_Area_Key, Object_Key
  Select * from DW_Object_Contents
  Select * from DW_Attributes
END

IF @Debug = 'False' SET NOCOUNT OFF
